package ru.dbotthepony.mc.otm.network

import io.netty.buffer.ByteBuf
import net.minecraft.core.UUIDUtil
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.RegistryFriendlyByteBuf
import net.minecraft.network.codec.ByteBufCodecs
import net.minecraft.network.codec.StreamCodec
import net.minecraft.resources.ResourceLocation
import net.minecraft.util.valueproviders.FloatProvider
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.mc.otm.core.math.readDecimal
import ru.dbotthepony.mc.otm.core.math.writeDecimal
import ru.dbotthepony.mc.otm.core.readItemType
import ru.dbotthepony.mc.otm.core.writeItemType
import kotlin.reflect.KMutableProperty1

object StreamCodecs {
	val NOTHING: MatteryStreamCodec<ByteBuf, Nothing?> = StreamCodec.of<ByteBuf, Nothing?>({ _, _ -> }, { null }).wrap()

	val BYTE: MatteryStreamCodec<ByteBuf, Byte> = StreamCodec.of({ s, v -> s.writeByte(v.toInt()) }, ByteBuf::readByte).wrap()
	val SHORT = ByteBufCodecs.SHORT.wrap()
	val INT = ByteBufCodecs.INT.wrap()
	val VAR_INT = ByteBufCodecs.VAR_INT.wrap()
	val LONG = StreamCodec.of(ByteBuf::writeLong, ByteBuf::readLong).wrap()
	val VAR_LONG = ByteBufCodecs.VAR_LONG.wrap()
	val DOUBLE = ByteBufCodecs.DOUBLE.wrap()
	val FLOAT = ByteBufCodecs.FLOAT.wrap()
	val BOOLEAN = ByteBufCodecs.BOOL.wrap()
	val STRING = ByteBufCodecs.STRING_UTF8.wrap()
	val UUID = UUIDUtil.STREAM_CODEC.wrap()
	val RESOURCE_LOCATION = ResourceLocation.STREAM_CODEC.wrap()

	val RGBA: MatteryStreamCodec<ByteBuf, RGBAColor> = StreamCodec.of<ByteBuf, RGBAColor>(
		{ s, v -> s.writeFloat(v.red); s.writeFloat(v.green); s.writeFloat(v.blue); s.writeFloat(v.alpha) },
		{ s -> RGBAColor(s.readFloat(), s.readFloat(), s.readFloat(), s.readFloat()) }).wrap()

	val ITEM_TYPE = StreamCodec.of(FriendlyByteBuf::writeItemType, FriendlyByteBuf::readItemType).wrap()
	val ITEM_TYPE_NULLABLE = ITEM_TYPE.nullable()
	val DECIMAL = StreamCodec.of(FriendlyByteBuf::writeDecimal, FriendlyByteBuf::readDecimal).wrap()
}
